import tempfile
import textwrap

import pytest


@pytest.fixture
def temp_config():
    with tempfile.NamedTemporaryFile() as config:
        config.write(textwrap.dedent('''
            some-secret: foo secret
        ''').encode('utf-8'))
        config.flush()
        yield config.name



import email.message
from email.mime.multipart import MIMEMultipart

import pytest

from laim.log import format_message_structure, format_context


@pytest.mark.parametrize('context,expected', [
    ({'somekey': 1}, 'somekey=1'),
    ({'somekey': 0}, 'somekey=0'),
    ({'empty': None}, 'empty='),
    ({'true': True}, 'true=true'),
    ({'false': False}, 'false=false'),
    ({'float': 1.23456789}, 'float=1.2346'),
    ({'mykey': 'my custom value'}, 'mykey="my custom value"'),
])
def test_logfmt_formatting(context, expected):
    assert format_context(context) == expected


def test_format_message_structure_plain():
    msg = email.message.Message()
    assert format_message_structure(msg) == 'text/plain'

    msg.set_type('text/html')
    assert format_message_structure(msg) == 'text/html'


def test_format_message_structure_multipart():
    msg = MIMEMultipart('alternative')
    attach1 = email.message.Message()
    attach1.set_type('text/plain')
    msg.attach(attach1)

    attach2 = email.message.Message()
    attach2.set_type('text/html')
    msg.attach(attach2)

    assert format_message_structure(msg) == 'multipart/alternative(text/plain, text/html)'


def test_format_message_structure_nested():
    msg = MIMEMultipart()
    attach1 = email.message.Message()
    attach1.set_type('text/plain')
    msg.attach(attach1)

    nested_msg = MIMEMultipart('alternative')
    nested1 = email.message.Message()
    nested1.set_type('image/jpg')
    nested_msg.attach(nested1)

    nested2 = email.message.Message()
    nested2.set_type('image/png')
    nested_msg.attach(nested2)

    msg.attach(nested_msg)

    expected = 'multipart/mixed(text/plain, multipart/alternative(image/jpg, image/png))'
    assert format_message_structure(msg) == expected



import os
import pwd
import re


def drop_privileges(user):
    pwd_details = pwd.getpwnam(user)

    # Remove group privileges
    os.setgroups([])

    os.setgid(pwd_details.pw_gid)
    os.setuid(pwd_details.pw_uid)


def unfold(folded):
    '''Helper to unfold headers'''
    if folded is None:
        return None
    return re.sub(r'\r?\n ', ' ', folded)



import asyncio
from queue import Queue
from unittest import mock

import pytest

from laim import Laim, unfold
from laim.laim import LaimHandler


def test_drops_privileges(temp_config):
    with mock.patch('laim.util.os') as os_mock:
        with mock.patch('pwd.getpwnam') as getpwnam_mock:
            with mock.patch('laim.laim.LaimController'):
                getpwnam_mock.return_value.pw_uid = 1337
                getpwnam_mock.return_value.pw_gid = 2448
                Laim(port=2525, config_file=temp_config)

    getpwnam_mock.assert_called_with('laim')
    os_mock.setgroups.assert_called_with([])
    os_mock.setuid.assert_called_with(1337)
    os_mock.setgid.assert_called_with(2448)


def test_loads_config(temp_config):
    with mock.patch('laim.laim.drop_privileges'):
        with mock.patch('laim.laim.LaimController'):
            handler = Laim(config_file=temp_config)
    assert handler.config['some-secret'] == 'foo secret'


def test_full_queue():
    queue = Queue(2)
    handler = LaimHandler(queue)
    envelope = mock.MagicMock()
    envelope.rcpt_tos = ['foo']
    loop = asyncio.get_event_loop()
    run = loop.run_until_complete
    def add_to_queue():
        return run(handler.handle_DATA(None, mock.Mock(), envelope))

    add_to_queue() == '250 OK'
    add_to_queue() == '250 OK'
    add_to_queue() == '552 Exceeded storage allocation'


@pytest.mark.parametrize('testcase', [
    ('foo\r\n bar', 'foo bar'),
    ('foo\n bar', 'foo bar'),
    ('foo\r\n bar\n baz', 'foo bar baz'),
    ('foo\r\n  bar', 'foo  bar'),
    (None, None), # convenient to avoid crashing if a header is missing
])
def test_unfold(testcase):
    folded, expected = testcase
    assert unfold(folded) == expected



import platform
import queue
import signal
import threading
import time
from collections import namedtuple
from email import message_from_string
from email.header import decode_header, make_header

import sdnotify
import setproctitle
import yaml
from aiosmtpd.controller import Controller
from aiosmtpd.smtp import SMTP

from .util import drop_privileges, unfold
from .log import log, format_message_structure
from ._version import __version__


TaskArguments = namedtuple('TaskArguments', 'sender recipients data')


class Laim:

    def __init__(
            self,
            port=25,
            user='laim',
            max_queue_size=50,
            config_file='/etc/laim/config.yml',
            smtp_kwargs=None,
    ):
        setproctitle.setproctitle('laim')
        self.queue = queue.Queue(max_queue_size)
        self.notifier = sdnotify.SystemdNotifier()
        handler = LaimHandler(self.queue)
        privilege_event = threading.Event()
        self.controller = LaimController(
            privilege_event,
            handler,
            port=port,
            smtp_kwargs=smtp_kwargs,
        )

        # Start the controller while we have the privileges to bind the port
        self.controller.start()

        # Let systemd we're done binding to the network socket
        self.notifier.notify('READY=1')
        log({
            'action': 'started',
            'port': port,
            'max_queue_size': max_queue_size,
            'config_file': config_file,
            'py': platform.python_version(),
        }, sender=self)

        with open(config_file, 'r') as config_fh:
            drop_privileges(user)
            self.config = yaml.safe_load(config_fh)

        self.stop_event = threading.Event()
        signal.signal(signal.SIGTERM, self._signalhandler)
        signal.signal(signal.SIGINT, self._signalhandler)

        # Allow client sessions to be created
        privilege_event.set()


    def handle_message(self, sender, recipients, message):
        raise NotImplementedError('The handler must implement handle_message')


    def run(self):
        worker_thread = threading.Thread(
            target=self._start_worker,
            name='Laim worker',
            daemon=False,
        )
        worker_thread.start()

        self.stop_event.wait()
        self.controller.stop()


    def stop(self):
        self.queue.put(None)
        self.stop_event.set()


    def _signalhandler(self, signum, frame): # pylint: disable=unused-argument
        self.stop()


    def _start_worker(self):
        while True:
            task_args = self.queue.get()
            if task_args is None:
                break

            start_time = time.time()
            message_string = task_args.data.decode('utf-8')
            message = message_from_string(message_string)
            raw_subject = message.get('subject')

            # Decode the subject to make it easier to consume for handlers
            decoded_subject = None
            if raw_subject:
                decoded_subject = str(make_header(decode_header(raw_subject)))
                message.replace_header('subject', decoded_subject)

            log_data = {
                'action': 'handle-message',
                'parse_time': '%.3fs' % (time.time() - start_time),
                'sender': task_args.sender,
                'recipients': ','.join(task_args.recipients),
                'msg_structure': format_message_structure(message),
                'raw_subject': unfold(raw_subject),
                'decoded_subject': decoded_subject,
                'msg_defects': ','.join(e.__class__.__name__ for e in message.defects),
            }

            try:
                handler_data = self.handle_message(task_args.sender, task_args.recipients, message)
                if handler_data:
                    log_data.update(handler_data)
            except Exception as ex: # pylint: disable=broad-except
                log_data['action'] = 'handle-message-error'
                log_data['error'] = ex.__class__.__name__
                log_data['error_msg'] = str(ex)
            finally:
                log(log_data, start_time, sender=self)


class LaimHandler:
    def __init__(self, task_queue):
        self.task_queue = task_queue


    async def handle_DATA(self, server, session, envelope): # pylint: disable=invalid-name,unused-argument
        start_time = time.time()
        mail_from = envelope.mail_from
        data = envelope.content
        recipients = envelope.rcpt_tos
        log_data = {
            'action': 'queued-message',
            'mail_from': mail_from,
            'recipients': ','.join(recipients),
            'msg_size': len(data),
            'client': session.host_name,
        }
        try:
            self.task_queue.put_nowait(TaskArguments(mail_from, recipients, data))
            log(log_data, start_time, sender=self)
        except queue.Full:
            log_data['action'] = 'queue-full'
            log(log_data, start_time, sender=self)
            return '552 Exceeded storage allocation'

        return '250 OK'


class LaimController(Controller):
    def __init__(self, privilege_event, handler, port, smtp_kwargs=None):
        super().__init__(handler, hostname='localhost', port=port)
        self.smtp_kwargs = smtp_kwargs
        self.privilege_event = privilege_event


    def factory(self):
        kwargs = {
            'enable_SMTPUTF8': True,
            'ident': 'laim %s' % __version__,
        }
        if self.smtp_kwargs:
            kwargs.update(self.smtp_kwargs)

        return LaimSMTP(self.privilege_event, self.handler, **kwargs)


class LaimSMTP(SMTP):
    '''
    Subclass to make sure no sessions are created before we've dropped
    privileges.
    '''
    def __init__(self, privilege_event, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.privilege_event = privilege_event


    def _create_session(self):
        self.privilege_event.wait()
        return super()._create_session()



import email.parser
import email.policy
import os
import pwd
import textwrap
from unittest import mock

import pytest

from laim.__main__ import read_message, main, extract_recipients_from_to_header

def test_read_message():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'From: foo\n',
        b'\n',
        b'Foo content\n',
        b'.\n',
        b'Should be ignored\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        message = read_message(stop_on_dot=True)

    assert message.get_payload() == 'Foo content\n'


def test_read_message_ignore_dot():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'From: foo\n',
        b'\n',
        b'Foo content\n',
        b'.\n',
        b'Should be included\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        message = read_message(stop_on_dot=False)

    assert message.get_payload() == 'Foo content\n.\nShould be included\n'


def test_message_parsing():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'From: foo\n',
        b'To: bar\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main([])

    assert_send_mail_called(send_mail_mock, None, [], textwrap.dedent('''\
        From: foo\r
        To: bar\r
        \r
        Message\r
    '''))


def test_message_without_from_and_with_F_argument():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'To: bar\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main(['-F', 'foo name'])

    current_user = pwd.getpwuid(os.getuid()).pw_name
    assert_send_mail_called(send_mail_mock, current_user, None, textwrap.dedent('''\
        To: bar\r
        From: "foo name" <%s>\r
        \r
        Message\r
    ''' % current_user))


def test_message_without_from_and_with_F_argument():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'To: bar\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main(['-F', 'foo name'])

    current_user = pwd.getpwuid(os.getuid()).pw_name
    assert_send_mail_called(send_mail_mock, current_user, [], textwrap.dedent('''\
        To: bar\r
        From: "foo name" <%s>\r
        \r
        Message\r
    ''' % current_user))


def test_message_without_to():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'Subject: test subject\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            with pytest.raises(SystemExit):
                main([])


def test_message_without_to_but_with_recipients():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'Subject: test subject\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main(['foo'])

    current_user = pwd.getpwuid(os.getuid()).pw_name
    assert_send_mail_called(send_mail_mock, current_user, ['foo'], textwrap.dedent('''\
        Subject: test subject\r
        From: %s\r
        \r
        Message\r
    ''' % current_user))


@pytest.mark.parametrize('flag', ('-r', '-f'))
def test_message_without_from_and_with_sender_argument(flag):
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'To: bar\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main([flag, 'foo'])

    assert_send_mail_called(send_mail_mock, 'foo', [], textwrap.dedent('''\
        To: bar\r
        From: foo\r
        \r
        Message\r
    '''))


def test_message_extract_recipient():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'From: foo\n',
        b'To: bar, foo\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main(['-t', 'baz', 'foo'])

    assert_send_mail_called(send_mail_mock, None, ['baz', 'foo', 'bar'], textwrap.dedent('''\
        From: foo\r
        To: bar, foo\r
        \r
        Message\r
    '''))


def test_quoted_printable_long_subject():
    stdin_mock = mock.Mock()
    stdin_mock.buffer = [
        b'Subject: =?utf-8?q?mail_with_long_long_long_long_long_long_long_long_long_long_long_s?=\n',
        b' =?utf-8?q?ubject?=\n',
        b'To: foo\n',
        b'\n',
        b'Message\n',
    ]
    with mock.patch('laim.__main__.sys.stdin', stdin_mock):
        with mock.patch('laim.__main__.send_mail') as send_mail_mock:
            main([])

    current_user = pwd.getpwuid(os.getuid()).pw_name
    assert_send_mail_called(send_mail_mock, current_user, [], textwrap.dedent('''\
        Subject: =?utf-8?q?mail_with_long_long_long_long_long_long_long_long_long_long_long_s?=\r
         =?utf-8?q?ubject?=\r
        To: foo\r
        From: %s\r
        \r
        Message\r
    ''' % current_user))


def assert_send_mail_called(send_mail_mock, sender, recipients, message_string):
    send_mail_mock.assert_called_with(sender, recipients, mock.ANY)
    sent_message = send_mail_mock.call_args[0][2]
    assert sent_message.as_string(policy=email.policy.SMTP) == message_string


@pytest.mark.parametrize('test_case', [
    ('foo', ['foo']),
    ('<foo>', ['foo']),
    ('"Foo Name" <foo@bar.com>, bar, Baz <bzz@baz.org>', ['foo@bar.com', 'bar', 'bzz@baz.org'])
])
def test_extract_recipients_from_to_header(test_case):
    to_header, expected = test_case
    actual = extract_recipients_from_to_header(to_header)
    assert actual == expected



'''
laim to sendmail compatibility interface.
'''

import argparse
import email.parser
import logging
import os
import pwd
import smtplib
import sys
import traceback
from email.headerregistry import AddressHeader

from laim._version import __version__

# Extracted as a constant to make it easy to override for tests
SMTP_PORT = 25


_logger = logging.getLogger('laim')


def main(argv=None):
    args = parse_args(argv)
    try:
        sendmail(args)
    except Exception as ex: # pylint: disable=broad-except
        _logger.debug(traceback.format_exc())
        _logger.warning('Failed to send mail: %s', ex)
        sys.exit(1)


def mailq(prog='mailq'):
    _logger.warning('%s: Mail queue is empty (and not used by laim)', prog)


def newaliases(prog='newaliases'):
    _logger.warning('%s: Aliases are not used by laim', prog)


def sendmail(args):
    if args.bp:
        mailq('sendmail')
        return

    if args.bi:
        newaliases('sendmail')
        return

    message = read_message(args.i)
    sender = None
    recipients = args.recipients
    if 'From' not in message:
        sender = args.r or pwd.getpwuid(os.getuid()).pw_name
        if args.F:
            message['From'] = '"%s" <%s>' % (args.F, sender)
        else:
            message['From'] = sender

    if 'To' not in message and not recipients:
        raise ValueError("Message doesn't have a To header and no recipients "
            'given on the command line')

    if args.t:
        recipients.extend(extract_recipients_from_to_header(message['To']))

    send_mail(sender, unique(recipients), message)


def extract_recipients_from_to_header(to_header):
    recipients = []
    address_header = AddressHeader.value_parser(to_header)
    for address in address_header:
        if address.token_type != 'address':
            continue
        for mailbox in address.all_mailboxes:
            recipients.append(mailbox.addr_spec)
    return recipients


def unique(iterable):
    seen = set()
    ret = []
    for item in iterable:
        if not item in seen:
            ret.append(item)
            seen.add(item)
    return ret


def read_message(stop_on_dot):
    binary_lines = []
    for line in sys.stdin.buffer:
        if stop_on_dot and line == b'.\n':
            break
        binary_lines.append(line)
    parser = email.parser.BytesParser()
    message = parser.parsebytes(b''.join(binary_lines))
    return message


def send_mail(sender, recipients, message):
    with smtplib.SMTP('127.0.0.1', port=SMTP_PORT, timeout=3) as smtp:
        smtp.send_message(message, from_addr=sender, to_addrs=recipients or None)


def parse_args(argv=None):
    parser = argparse.ArgumentParser(description=__doc__, add_help=False)
    parser.add_argument('recipients', nargs='*')
    parser.add_argument('-bp', '-q', action='store_true',
        help='List the mail queue')
    parser.add_argument('-bi', '-I', action='store_true',
        help='Initialize the alias database. Ignored.')
    parser.add_argument('-i', '-oi', action='store_false', default=True,
        help="When reading a message from standard input, don't treat a "
        "line with only a .  character  as the end of input.")
    parser.add_argument('-F', help='Set the full name of the sender. Only '
        'used for messages without a From: header')
    parser.add_argument('-r', '-f', help='Set the envelope sender.')
    parser.add_argument('-t', action='store_true', help='Extract recipients '
        'from message headers. These are added to any recipients specified on '
        'the command line.')
    parser.add_argument('-v', action='store_true', help='More verbose logging')

    args, unparsed = parser.parse_known_args(argv)

    configure_logger(logging.DEBUG if args.v else logging.INFO)
    _logger.debug('laim version %s', __version__)

    if unparsed:
        _logger.debug('laim: ignoring the following arguments: %s',
            ' '.join(unparsed))

    return args


def configure_logger(level):
    handler = logging.StreamHandler()
    handler.setLevel(level)
    _logger.addHandler(handler)
    _logger.setLevel(level)


if __name__ == '__main__':
    main() # pragma: no cover



import threading
import time
from unittest import mock

import pytest

from laim import Laim
from laim.__main__ import main

pytestmark = pytest.mark.integration


def test_sendmail_delivery_to_handler(temp_config):
    handled_event = threading.Event()
    lock = threading.Lock()
    received_sender = None
    received_recipients = None
    received_message = None

    class Handler(Laim):
        def handle_message(self, sender, recipients, message):
            nonlocal received_sender, received_recipients, received_message
            with lock:
                received_sender = sender
                received_recipients = recipients
                received_message = message
            handled_event.set()
            self.stop()

    with mock.patch('laim.laim.drop_privileges'):
        handler = Handler(port=2525, config_file=temp_config)

    def start_handler():
        handler.run()
    handler_thread = threading.Thread(target=start_handler)
    handler_thread.start()

    # Wait for the handler to start and bind to the port
    time.sleep(1)

    with mock.patch('laim.__main__.SMTP_PORT', 2525):
        stdin_mock = mock.Mock()
        stdin_mock.buffer = [
            b'From: foo@bar.com\r\n',
            b'To: bar@foo.com\r\n',
            b'Subject: =?utf-8?q?subject?=\r\n',
            b'\r\n',
            b'Hello, laim!\r\n',
        ]
        with mock.patch('laim.__main__.sys.stdin', stdin_mock):
            try:
                main([])
            except:
                handler.stop()
                pytest.fail()

    handled_event.wait(2)

    assert handled_event.is_set()

    with lock:
        assert received_sender == 'foo@bar.com'
        assert received_recipients == ['bar@foo.com']
        assert received_message.get_payload() == 'Hello, laim!\r\n'
        assert received_message.get('subject') == 'subject'



from unittest import mock

import pytest

from laim import Laim
from laim.laim import TaskArguments

pytestmark = pytest.mark.integration


def test_handler(temp_config):
    class Handler(Laim):
        def handle_message(self, sender, recipients, message):
            self.stop()

    with mock.patch('laim.laim.drop_privileges'):
        with mock.patch('laim.laim.LaimController'):
            handler = Handler(config_file=temp_config)

    handler.queue.put(TaskArguments(mock.Mock(), [], b''))
    handler._start_worker()

    assert handler.stop_event.is_set()


def test_crashing_handler(temp_config):
    class Handler(Laim):
        def handle_message(self, sender, recipients, message):
            self.stop()
            raise ValueError()

    with mock.patch('laim.laim.drop_privileges'):
        with mock.patch('laim.laim.LaimController'):
            handler = Handler(config_file=temp_config)

    handler.queue.put(TaskArguments(mock.Mock(), [], b''))
    handler._start_worker()

    # the worker thread shouldn't crash
    assert handler.stop_event.is_set()



import numbers
import re
import threading
import time

from blinker import signal

from ._version import __version__

STDOUT_LOCK = threading.Lock()
NEEDS_QUOTES_RE = re.compile(r'[\s=]')

before_log = signal('before-log')


def log(context, start_time=None, sender=None):
    context['version'] = __version__

    if start_time is not None:
        context['duration_ms'] = (time.time() - start_time)*1000

    before_log.send(sender, log_data=context)
    message = format_context(context)

    with STDOUT_LOCK:
        print(message)


def format_context(context):
    return ' '.join(
        format_key_value_pair(key, val) for (key, val) in context.items()
    )


def format_key_value_pair(key, value):
    if value is None:
        value = ''
    elif value is True:
        value = 'true'
    elif value is False:
        value = 'false'
    elif isinstance(value, numbers.Integral):
        value = str(value)
    elif isinstance(value, numbers.Real):
        value = '%.4f' % value
    else:
        value = str(value)

    should_quote = NEEDS_QUOTES_RE.search(value)

    if should_quote:
        value = '"%s"' % value

    return '%s=%s' % (key, value)


def format_message_structure(message):
    if not message.is_multipart():
        return message.get_content_type()

    ret = []
    ret.append(message.get_content_type())
    ret.append('(')

    first = True
    for part in message.get_payload():
        if not first:
            ret.append(', ')
        ret.append(format_message_structure(part))
        first = False

    ret.append(')')

    return ''.join(ret)
